# type: ignore
__author__ = "o-murphy"
__copyright__ = (
    "Copyright 2023 Dmytro Yaroshenko (https://github.com/o-murphy)",
    "Copyright 2024 David Bookstaber (https://github.com/dbookstaber)"
)

__credits__ = ["o-murphy", "dbookstaber"]

from .trajectory_calc import EulerTrajectoryCalc
from .vector import Vector
from .trajectory_data import TrajectoryData
from .interface import EulerCalculator

__all__ = (
    'EulerTrajectoryCalc',
    'Vector',
    'TrajectoryData',
    'EulerCalculator'
)



"""Unittests for the py_ballisticcalc library"""

import unittest
import copy
from py_ballisticcalc import (
    DragModel, Ammo, Weapon, Shot, Wind, Atmo, TableG7, RangeError,
)
from py_ballisticcalc.unit import *
from CyEulerBallistic import EulerCalculator


class TestComputer(unittest.TestCase):
    """Basic verifications that wind, spin, and cant values produce effects of correct sign and magnitude"""

    def setUp(self):
        """Baseline shot has barrel at zero elevation"""
        self.range = 1000
        self.step = 100
        self.dm = DragModel(0.22, TableG7, 168, 0.308, 1.22)
        self.ammo = Ammo(self.dm, Velocity.FPS(2600))
        self.weapon = Weapon(4, 12)
        self.atmosphere = Atmo.icao()  # Standard sea-level atmosphere
        self.calc = EulerCalculator()
        self.baseline_shot = Shot(weapon=self.weapon, ammo=self.ammo, atmo=self.atmosphere)
        self.baseline_trajectory = self.calc.fire(shot=self.baseline_shot, trajectory_range=self.range,
                                                  trajectory_step=self.step)

    # region Cant_angle
    def test_cant_zero_elevation(self):
        """Cant_angle = 90 degrees with zero barrel elevation should match baseline with:
            drop+=sight_height, windage-=sight_height
        """
        canted = copy.copy(self.baseline_shot)
        canted.cant_angle = Angular.Degree(90)
        t = self.calc.fire(canted, trajectory_range=self.range, trajectory_step=self.step)
        self.assertAlmostEqual(t.trajectory[5].height.raw_value - self.weapon.sight_height.raw_value,
                               self.baseline_trajectory[5].height.raw_value)
        self.assertAlmostEqual(t.trajectory[5].windage.raw_value + self.weapon.sight_height.raw_value,
                               self.baseline_trajectory[5].windage.raw_value)

    def test_cant_positive_elevation(self):
        """Cant_angle = 90 degrees with positive barrel elevation and zero twist should match baseline with:
            drop+=sight_height, windage-=sight_height at muzzle, increasingly positive down-range
        """
        canted = Shot(weapon=Weapon(sight_height=self.weapon.sight_height, twist=0, zero_elevation=Angular.Mil(2)),
                      ammo=self.ammo, atmo=self.atmosphere, cant_angle=Angular.Degree(90))
        t = self.calc.fire(canted, trajectory_range=self.range, trajectory_step=self.step)
        self.assertAlmostEqual(t.trajectory[5].height.raw_value - self.weapon.sight_height.raw_value,
                               self.baseline_trajectory[5].height.raw_value, 2)
        self.assertAlmostEqual(t.trajectory[0].windage.raw_value, -self.weapon.sight_height.raw_value)
        self.assertGreater(t.trajectory[5].windage.raw_value, t.trajectory[3].windage.raw_value)

    def test_cant_zero_sight_height(self):
        """Cant_angle = 90 degrees with sight_height=0 and barrel_elevation=0 should match baseline with:
            drop+=baseline.sight_height, windage no change
        """
        canted = Shot(weapon=Weapon(sight_height=0, twist=self.weapon.twist),
                      ammo=self.ammo, atmo=self.atmosphere, cant_angle=Angular.Degree(90))
        t = self.calc.fire(canted, trajectory_range=self.range, trajectory_step=self.step)
        self.assertAlmostEqual(t.trajectory[5].height.raw_value - self.weapon.sight_height.raw_value,
                               self.baseline_trajectory[5].height.raw_value)
        self.assertAlmostEqual(t.trajectory[5].windage, self.baseline_trajectory[5].windage)

    # endregion Cant_angle

    # region Wind
    def test_wind_from_left(self):
        """Wind from left should increase windage"""
        shot = Shot(weapon=self.weapon, ammo=self.ammo, atmo=self.atmosphere,
                    winds=[Wind(Velocity(5, Velocity.MPH), Angular(3, Angular.OClock))])
        t = self.calc.fire(shot, trajectory_range=self.range, trajectory_step=self.step)
        self.assertGreater(t.trajectory[5].windage, self.baseline_trajectory[5].windage)

    def test_wind_from_right(self):
        """Wind from right should decrease windage"""
        shot = Shot(weapon=self.weapon, ammo=self.ammo, atmo=self.atmosphere,
                    winds=[Wind(Velocity(5, Velocity.MPH), Angular(9, Angular.OClock))])
        t = self.calc.fire(shot, trajectory_range=self.range, trajectory_step=self.step)
        self.assertLess(t.trajectory[5].windage, self.baseline_trajectory[5].windage)

    def test_wind_from_back(self):
        """Wind from behind should decrease drop"""
        shot = Shot(weapon=self.weapon, ammo=self.ammo, atmo=self.atmosphere,
                    winds=[Wind(Velocity(5, Velocity.MPH), Angular(0, Angular.OClock))])
        t = self.calc.fire(shot, trajectory_range=self.range, trajectory_step=self.step)
        self.assertGreater(t.trajectory[5].height, self.baseline_trajectory[5].height)

    def test_wind_from_front(self):
        """Wind from in front should increase drop"""
        shot = Shot(weapon=self.weapon, ammo=self.ammo, atmo=self.atmosphere,
                    winds=[Wind(Velocity(5, Velocity.MPH), Angular(6, Angular.OClock))])
        t = self.calc.fire(shot, trajectory_range=self.range, trajectory_step=self.step)
        self.assertLess(t.trajectory[5].height, self.baseline_trajectory[5].height)

    def test_multiple_wind(self):
        shot = Shot(weapon=self.weapon, ammo=self.ammo, atmo=self.atmosphere,
                    winds=[Wind(Velocity.MPS(4), Angular.OClock(9), until_distance=Distance.Meter(500)),
                           Wind(Velocity.MPS(4), Angular.OClock(3), until_distance=Distance.Meter(800))])
        t = self.calc.fire(shot, trajectory_range=self.range, trajectory_step=self.step)
        self.assertLess(t.trajectory[5].windage, self.baseline_trajectory[5].windage)

    def test_no_winds(self):
        shot = Shot(weapon=self.weapon, ammo=self.ammo, atmo=self.atmosphere,
                    winds=[])
        # set empty list
        shot.winds = []
        try:
            self.calc.fire(shot, trajectory_range=self.range, trajectory_step=self.step)
        except Exception as e:
            self.fail("self.calc.fire() raised ExceptionType unexpectedly!")

        self.winds = None
        try:
            self.calc.fire(shot, trajectory_range=self.range, trajectory_step=self.step)
        except Exception as e:
            self.fail("self.calc.fire() raised ExceptionType unexpectedly!")

    # endregion Wind

    # region Twist
    def test_no_twist(self):
        """Barrel with no twist should have no spin drift"""
        shot = Shot(weapon=Weapon(twist=0), ammo=self.ammo, atmo=self.atmosphere)
        t = self.calc.fire(shot, trajectory_range=self.range, trajectory_step=self.step)
        self.assertEqual(t.trajectory[5].windage.raw_value, 0)

    def test_twist(self):
        """Barrel with right-hand twist should have positive spin drift.
            Barrel with left-hand twist should have negative spin drift.
            Faster twist rates should produce larger drift.
        """
        shot = Shot(weapon=Weapon(twist=12), ammo=self.ammo, atmo=self.atmosphere)
        twist_right = self.calc.fire(shot, trajectory_range=self.range, trajectory_step=self.step)
        self.assertGreater(twist_right.trajectory[5].windage.raw_value, 0)
        shot = Shot(weapon=Weapon(twist=-8), ammo=self.ammo, atmo=self.atmosphere)
        twist_left = self.calc.fire(shot, trajectory_range=self.range, trajectory_step=self.step)
        self.assertLess(twist_left.trajectory[5].windage.raw_value, 0)
        # Faster twist should produce larger drift:
        self.assertGreater(-twist_left.trajectory[5].windage.raw_value, twist_right.trajectory[5].windage.raw_value)

    # endregion Twist

    # region Atmo
    def test_humidity(self):
        """Increasing relative humidity should decrease drop (due to decreasing density)"""
        humid = Atmo(humidity=.9)  # 90% humidity
        shot = Shot(weapon=self.weapon, ammo=self.ammo, atmo=humid)
        t = self.calc.fire(shot=shot, trajectory_range=self.range, trajectory_step=self.step)
        self.assertGreater(t.trajectory[5].height, self.baseline_trajectory[5].height)

    def test_temp_atmo(self):
        """Dropping temperature should increase drop (due to increasing density)"""
        cold = Atmo(temperature=Temperature.Celsius(0))
        shot = Shot(weapon=self.weapon, ammo=self.ammo, atmo=cold)
        t = self.calc.fire(shot=shot, trajectory_range=self.range, trajectory_step=self.step)
        self.assertLess(t.trajectory[5].height, self.baseline_trajectory[5].height)

    def test_altitude(self):
        """Increasing altitude should decrease drop (due to decreasing density)"""
        high = Atmo.icao(Distance.Foot(5000))
        shot = Shot(weapon=self.weapon, ammo=self.ammo, atmo=high)
        t = self.calc.fire(shot=shot, trajectory_range=self.range, trajectory_step=self.step)
        self.assertGreater(t.trajectory[5].height, self.baseline_trajectory[5].height)

    def test_pressure(self):
        """Decreasing pressure should decrease drop (due to decreasing density)"""
        thin = Atmo(pressure=Pressure.InHg(20.0))
        shot = Shot(weapon=self.weapon, ammo=self.ammo, atmo=thin)
        t = self.calc.fire(shot=shot, trajectory_range=self.range, trajectory_step=self.step)
        self.assertGreater(t.trajectory[5].height, self.baseline_trajectory[5].height)

    # endregion Atmo

    # region Ammo
    def test_ammo_drag(self):
        """Increasing ballistic coefficient (bc) should decrease drop"""
        tdm = DragModel(self.dm.BC + 0.5, self.dm.drag_table, self.dm.weight, self.dm.diameter, self.dm.length)
        slick = Ammo(tdm, self.ammo.mv)
        shot = Shot(weapon=self.weapon, ammo=slick, atmo=self.atmosphere)
        t = self.calc.fire(shot=shot, trajectory_range=self.range, trajectory_step=self.step)
        self.assertGreater(t.trajectory[5].height, self.baseline_trajectory[5].height)

    def test_ammo_optional(self):
        """DragModel.weight and .diameter, and Ammo.length, are only relevant when computing
            spin-drift.  Drop should match baseline with those parameters omitted.
        """
        tdm = DragModel(self.dm.BC, self.dm.drag_table)
        tammo = Ammo(tdm, mv=self.ammo.mv)
        shot = Shot(weapon=self.weapon, ammo=tammo, atmo=self.atmosphere)
        t = self.calc.fire(shot=shot, trajectory_range=self.range, trajectory_step=self.step)
        self.assertEqual(t.trajectory[5].height, self.baseline_trajectory[5].height)

    def test_powder_sensitivity(self):
        """With _globalUsePowderSensitivity: Reducing temperature should reduce muzzle velocity"""
        self.ammo.calc_powder_sens(Velocity.FPS(2550), Temperature.Celsius(0))

        with self.subTest("don't uses powder sensitivity"):
            cold = Atmo(temperature=Temperature.Celsius(-5))
            shot = Shot(weapon=self.weapon, ammo=self.ammo, atmo=cold)
            t = self.calc.fire(shot=shot, trajectory_range=self.range, trajectory_step=self.step)
            self.assertEqual(t.trajectory[0].velocity, self.baseline_trajectory[0].velocity)

        self.ammo.use_powder_sensitivity = True

        with self.subTest("powder temperature the same as atmosphere temperature"):
            cold = Atmo(temperature=Temperature.Celsius(-5))
            shot = Shot(weapon=self.weapon, ammo=self.ammo, atmo=cold)
            t = self.calc.fire(shot=shot, trajectory_range=self.range, trajectory_step=self.step)
            self.assertLess(t.trajectory[0].velocity, self.baseline_trajectory[0].velocity)

        with self.subTest("different powder temperature"):
            cold = Atmo(powder_t=Temperature.Celsius(-5))
            shot = Shot(weapon=self.weapon, ammo=self.ammo, atmo=cold)
            t = self.calc.fire(shot=shot, trajectory_range=self.range, trajectory_step=self.step)
            self.assertLess(t.trajectory[0].velocity, self.baseline_trajectory[0].velocity)

        self.ammo.use_powder_sensitivity = False

    # @unittest.skip("Raises ZeroDivisionError")
    def test_zero_velocity(self):
        tdm = DragModel(self.dm.BC + 0.5, self.dm.drag_table, self.dm.weight, self.dm.diameter, self.dm.length)
        slick = Ammo(tdm, 0)
        shot = Shot(weapon=self.weapon, ammo=slick, atmo=self.atmosphere)

        with self.assertRaises(RangeError):
            self.calc.fire(shot=shot, trajectory_range=self.range, trajectory_step=self.step)

    # endregion Ammo

    # region Shot
    def test_winds_sort(self):
        winds = [
            Wind(Unit.MPS(0), Unit.Degree(90), Unit.Meter(100)),
            Wind(Unit.MPS(1), Unit.Degree(60), Unit.Meter(300)),
            Wind(Unit.MPS(2), Unit.Degree(30), Unit.Meter(200)),
            Wind(Unit.MPS(2), Unit.Degree(30), Unit.Meter(50)),
        ]

        # sorted_winds = sorted(winds, key=lambda winds: winds.until_distance.raw_value)

        # print()
        shot = Shot(
            None, None, 0, 0, 0, None,
            winds
        )
        sorted_winds = shot.winds
        self.assertIs(sorted_winds[0], winds[3])
        self.assertIs(sorted_winds[1], winds[0])
        self.assertIs(sorted_winds[2], winds[2])
        self.assertIs(sorted_winds[3], winds[1])


# endregion Shot


if __name__ == '__main__':
    unittest.main()



# type: ignore
from py_ballisticcalc.unit import Angular, Distance, Energy, Velocity, Weight
from typing_extensions import NamedTuple, Union, Tuple

__all__ = ['TrajectoryData']

class TrajectoryData(NamedTuple):
    __slots__: Tuple[str, ...]
    _field: Tuple[str, ...]

    time: float
    distance: Distance
    velocity: Velocity
    mach: float
    height: Distance
    target_drop: Distance
    drop_adj: Angular
    windage: Distance
    windage_adj: Angular
    look_distance: Distance
    angle: Angular
    density_factor: float
    drag: float
    energy: Energy
    ogw: Weight
    flag: Union[int]
    def formatted(self) -> tuple[str, ...]: ...
    def in_def_units(self) -> tuple[float, ...]: ...



# type: ignore
from dataclasses import dataclass
from typing_extensions import Union

__all__ = ['Vector']


@dataclass
class Vector:
    x: float
    y: float
    z: float
    def magnitude(self) -> float: ...
    def mul_by_const(self, a: float) -> Vector: ...
    def mul_by_vector(self, b: Vector) -> float: ...
    def add(self, b: Vector) -> Vector: ...
    def subtract(self, b: Vector) -> Vector: ...
    def negate(self) -> Vector: ...
    def normalize(self) -> Vector: ...
    def __add__(self, other: Vector) -> Vector: ...
    def __radd__(self, other: Vector) -> Vector: ...
    def __iadd__(self, other: Vector) -> Vector: ...
    def __sub__(self, other: Vector) -> Vector: ...
    def __rsub__(self, other: Vector) -> Vector: ...
    def __isub__(self, other: Vector) -> Vector: ...
    def __mul__(self, other: Union[int, float, 'Vector']) -> Union[float, 'Vector']: ...
    def __rmul__(self, other: Union[int, float, 'Vector']) -> Union[float, 'Vector']: ...
    def __imul__(self, other: Union[int, float, 'Vector']) -> Union[float, 'Vector']: ...
    def __neg__(self) -> Vector: ...
    def __init__(self, x: float, y: float, z: float) -> None: ...



"""Unittests for the py_ballisticcalc library"""

import unittest
from math import fabs
from py_ballisticcalc import *
from CyEulerBallistic import EulerCalculator


class TestTrajectory(unittest.TestCase):

    def test_zero1(self):
        dm = DragModel(0.365, TableG1, 69, 0.223, 0.9)
        ammo = Ammo(dm, 2600)
        weapon = Weapon(Distance(3.2, Distance.Inch))
        atmosphere = Atmo.icao()
        calc = EulerCalculator()
        zero_angle = calc.barrel_elevation_for_target(Shot(weapon=weapon, ammo=ammo, atmo=atmosphere),
                                     Distance(100, Distance.Yard))

        self.assertAlmostEqual(zero_angle >> Angular.Radian, 0.0016514, 6,
                               f'TestZero1 failed {zero_angle >> Angular.Radian:.10f}')

    def test_zero2(self):
        dm = DragModel(0.223, TableG7, 69, 0.223, 0.9)
        ammo = Ammo(dm, 2750)
        weapon = Weapon(Distance(2, Distance.Inch))
        atmosphere = Atmo.icao()
        calc = EulerCalculator()
        zero_angle = calc.barrel_elevation_for_target(Shot(weapon=weapon, ammo=ammo, atmo=atmosphere),
                                     Distance(100, Distance.Yard))

        self.assertAlmostEqual(zero_angle >> Angular.Radian, 0.0012286, 6,
                               f'TestZero2 failed {zero_angle >> Angular.Radian:.10f}')

    def custom_assert_equal(self, a, b, accuracy, name):
        with self.subTest(name=name):
            self.assertLess(fabs(a - b), accuracy, f'Equality {name} failed (|{a} - {b}|, {accuracy} digits)')

    def validate_one(self, data: TrajectoryData, distance: float, velocity: float,
                     mach: float, energy: float, path: float, hold: float,
                     windage: float, wind_adjustment: float, time: float, ogv: float,
                     adjustment_unit: Unit):

        self.custom_assert_equal(distance, data.distance >> Distance.Yard, 0.1, "Distance")
        self.custom_assert_equal(velocity, data.velocity >> Velocity.FPS, 5, "Velocity")
        self.custom_assert_equal(mach, data.mach, 0.005, "Mach")
        self.custom_assert_equal(energy, data.energy >> Energy.FootPound, 5, "Energy")
        self.custom_assert_equal(time, data.time, 0.06, "Time")
        self.custom_assert_equal(ogv, data.ogw >> Weight.Pound, 1, "OGV")

        if distance >= 800:
            self.custom_assert_equal(path, data.height >> Distance.Inch, 4, 'Drop')
        elif distance >= 500:
            self.custom_assert_equal(path, data.height >> Distance.Inch, 1, 'Drop')
        else:
            self.custom_assert_equal(path, data.height >> Distance.Inch, 0.5, 'Drop')

        if distance > 1:
            self.custom_assert_equal(hold, data.drop_adj >> adjustment_unit, 0.5, 'Hold')

        if distance >= 800:
            self.custom_assert_equal(windage, data.windage >> Distance.Inch, 1.5, "Windage")
        elif distance >= 500:
            self.custom_assert_equal(windage, data.windage >> Distance.Inch, 1, "Windage")
        else:
            self.custom_assert_equal(windage, data.windage >> Distance.Inch, 0.5, "Windage")

        if distance > 1:
            self.custom_assert_equal(wind_adjustment,
                                     data.windage_adj >> adjustment_unit, 0.5, "WAdj")

    def test_path_g1(self):
        dm = DragModel(0.223, TableG1, 168, 0.308, 1.282)
        ammo = Ammo(dm, Velocity(2750, Velocity.FPS))
        weapon = Weapon(Distance(2, Distance.Inch), zero_elevation=Angular(0.001228, Angular.Radian))
        atmosphere = Atmo.icao()
        shot_info = Shot(weapon=weapon, ammo=ammo, atmo=atmosphere,
                         winds=[Wind(Velocity(5, Velocity.MPH), Angular(10.5, Angular.OClock))])

        calc = EulerCalculator()
        data = calc.fire(shot_info, Distance.Yard(1000), Distance.Yard(100)).trajectory
        self.assertEqual(len(data), 11, "Trajectory Row Count")

        # Dist(yd), vel(fps), Mach, energy(ft-lb), drop(in), drop(mil), wind(in), wind(mil), time, ogw
        test_data = [
            [data[0], 0, 2750, 2.463, 2820.6, -2, 0, 0, 0, 0, 880, Angular.MOA],
            [data[1], 100, 2351.2, 2.106, 2061, 0, 0, -0.6, -0.6, 0.118, 550, Angular.MOA],
            [data[5], 500, 1169.1, 1.047, 509.8, -87.9, -16.8, -19.5, -3.7, 0.857, 67, Angular.MOA],
            [data[10], 1000, 776.4, 0.695, 224.9, -823.9, -78.7, -87.5, -8.4, 2.495, 20, Angular.MOA]
        ]

        failures = []

        for i, d in enumerate(test_data):
            try:
                with self.subTest(f"validate one {i}"):
                    self.validate_one(*d)
            except AssertionError as e:
                failures.append(f"Subtest {i} failed: {str(e)}")

        if failures:
            self.fail("\n".join(failures))  # Raise a single failure with all messages

    def test_path_g7(self):
        dm = DragModel(0.223, TableG7, 168, 0.308, 1.282)
        ammo = Ammo(dm, Velocity(2750, Velocity.FPS))
        weapon = Weapon(2, 12, zero_elevation=Angular.MOA(4.221))
        shot_info = Shot(weapon=weapon, ammo=ammo, winds=[Wind(Velocity(5, Velocity.MPH), Angular.Degree(-45))])

        calc = EulerCalculator()
        data = calc.fire(shot_info, Distance.Yard(1000), Distance.Yard(100)).trajectory
        self.assertEqual(len(data), 11, "Trajectory Row Count")

        # Dist(yd), vel(fps), Mach, energy(ft-lb), drop(in), drop(mil), wind(in), wind(mil), time, ogw
        test_data = [
            [data[0], 0,     2750, 2.46, 2821,  -2.0,   0.0,   0.0,  0.00, 0.000, 880, Angular.Mil],
            [data[1], 100,   2545, 2.28, 2416,   0.0,   0.0,  -0.2, -0.06, 0.113, 698, Angular.Mil],
            [data[5], 500,   1814, 1.62, 1227, -56.2,  -3.2,  -6.3, -0.36, 0.672, 252, Angular.Mil],
            [data[10], 1000, 1086, 0.97, 440, -399.9, -11.3, -31.6, -0.90, 1.748, 54, Angular.Mil]
        ]

        failures = []

        for i, d in enumerate(test_data):
            try:
                with self.subTest(f"validate one {i}"):
                    self.validate_one(*d)
            except AssertionError as e:
                failures.append(f"Subtest {i} failed: {str(e)}")

        if failures:
            self.fail("\n".join(failures))  # Raise a single failure with all messages


if __name__ == '__main__':
    unittest.main()



"Unit tests of multiple-bc drag models"

import unittest
from py_ballisticcalc import *

from CyEulerBallistic import EulerCalculator


class TestMBC(unittest.TestCase):

    def setUp(self) -> None:
        """Establish baseline trajectory"""
        self.range = 1000
        self.step = 100
        self.dm = DragModel(0.22, TableG7)
        self.ammo = Ammo(self.dm, Velocity.FPS(2600))
        self.weapon = Weapon(4, 12)
        self.calc = EulerCalculator()
        self.baseline_shot = Shot(weapon=self.weapon, ammo=self.ammo)
        self.baseline_trajectory = self.calc.fire(shot=self.baseline_shot, trajectory_range=self.range, trajectory_step=self.step).trajectory

    def test_mbc1(self):
        """We should get the same trajectory whether we give single bc or use multi-bc with single value"""
        dm_multi = DragModelMultiBC([BCPoint(.22, V=Velocity.FPS(2500)), BCPoint(.22, V=Velocity.FPS(1500)), BCPoint(BC=.22, Mach=3)], TableG7)
        multi_shot = Shot(weapon=self.weapon, ammo=Ammo(dm_multi, self.ammo.mv))
        multi_trajectory = self.calc.fire(shot=multi_shot, trajectory_range=self.range, trajectory_step=self.step).trajectory
        for i in range(len(multi_trajectory)):
            self.assertEqual(multi_trajectory[i].formatted(), self.baseline_trajectory[i].formatted())

    def test_mbc2(self):
        """Setting different bc above muzzle velocity should have no effect"""
        dm_multi = DragModelMultiBC([BCPoint(.22, V=Velocity.FPS(2700)), BCPoint(.5, V=Velocity.FPS(3500))], TableG7)
        multi_shot = Shot(weapon=self.weapon, ammo=Ammo(dm_multi, self.ammo.mv))
        multi_trajectory = self.calc.fire(shot=multi_shot, trajectory_range=self.range, trajectory_step=self.step).trajectory
        for i in range(len(multi_trajectory)):
            self.assertEqual(multi_trajectory[i].formatted(), self.baseline_trajectory[i].formatted())

    def test_mbc3(self):
        """Setting higher bc should result in higher downrange velocities"""
        # So here we'll boost the bc for velocities lower than the baseline's velocity at 200 yards
        dm_multi = DragModelMultiBC([BCPoint(.5, V=self.baseline_trajectory[3].velocity), BCPoint(.22, V=self.baseline_trajectory[2].velocity)], TableG7)
        multi_shot = Shot(weapon=self.weapon, ammo=Ammo(dm_multi, self.ammo.mv))
        multi_trajectory = self.calc.fire(shot=multi_shot, trajectory_range=self.range, trajectory_step=self.step).trajectory
        # Should show no change before 200 yards
        self.assertEqual(multi_trajectory[1].velocity.raw_value, self.baseline_trajectory[1].velocity.raw_value)
        # Should be faster at any point after 200 yards
        self.assertGreater(multi_trajectory[4].velocity.raw_value, self.baseline_trajectory[4].velocity.raw_value)

    def test_mbc(self):
        dm = DragModelMultiBC([BCPoint(0.275, V=Velocity.MPS(800)), BCPoint(0.255, V=Velocity.MPS(500)), BCPoint(0.26, V=Velocity.MPS(700))],
                              TableG7, weight=178, diameter=.308)
        self.assertAlmostEqual(dm.drag_table[0].CD, 0.1259323091692403)
        self.assertAlmostEqual(dm.drag_table[-1].CD, 0.1577125859466895)
    
    def test_mbc_valid(self):
        # Litz's multi-bc table comversion to CDM, 338LM 285GR HORNADY ELD-M
        dm = DragModelMultiBC([BCPoint(0.417, V=Velocity.MPS(745)), BCPoint(0.409, V=Velocity.MPS(662)), BCPoint(0.4, V=Velocity.MPS(580))],
                              drag_table=TableG7,
                              weight=Weight.Grain(285),
                              diameter=Distance.Inch(0.338)
                              )
        cds = [p.CD for p in dm.drag_table]
        machs = [p.Mach for p in dm.drag_table]
        reference = (
            (1, 0.3384895315),
            (2, 0.2585639866),
            (3, 0.2069547831),
            (4, 0.1652052415),
            (5, 0.1381406102),
        )
        for mach, cd in reference:
            idx = machs.index(mach)
            with self.subTest(mach=mach):
                self.assertAlmostEqual(cds[idx], cd, 3)



# type: ignore
"""Implements basic interface for the ballistics calculator"""
from dataclasses import dataclass, field

from typing_extensions import Union, List, Optional

# pylint: disable=import-error,no-name-in-module,wildcard-import
from CyEulerBallistic.trajectory_calc import EulerTrajectoryCalc
from py_ballisticcalc.conditions import Shot
from py_ballisticcalc.drag_model import DragDataPoint
from py_ballisticcalc.trajectory_data import HitResult
from py_ballisticcalc.unit import Angular, Distance, PreferredUnits
from py_ballisticcalc.interface_config import InterfaceConfigDict, create_interface_config


@dataclass
class EulerCalculator:
    """Basic interface for the ballistics calculator"""

    _config: Optional[InterfaceConfigDict] = field(default=None)
    _calc: EulerTrajectoryCalc = field(init=False, repr=False, compare=False)

    def __post_init__(self):
        self._calc = EulerTrajectoryCalc(create_interface_config(self._config))

    @property
    def cdm(self) -> List[DragDataPoint]:
        """returns custom drag function based on input data"""
        return self._calc.table_data

    def barrel_elevation_for_target(self, shot: Shot, target_distance: Union[float, Distance]) -> Angular:
        """Calculates barrel elevation to hit target at zero_distance.
        :param shot: Shot instance for which calculate barrel elevation is
        :param target_distance: Look-distance to "zero," which is point we want to hit.
            This is the distance that a rangefinder would return with no ballistic adjustment.
            NB: Some rangefinders offer an adjusted distance based on inclinometer measurement.
                However, without a complete ballistic model these can only approximate the effects
                on ballistic trajectory of shooting uphill or downhill.  Therefore:
                For maximum accuracy, use the raw sight distance and look_angle as inputs here.
        """
        target_distance = PreferredUnits.distance(target_distance)
        total_elevation = self._calc.zero_angle(shot, target_distance)
        return Angular.Radian(
            (total_elevation >> Angular.Radian) - (shot.look_angle >> Angular.Radian)
        )

    def set_weapon_zero(self, shot: Shot, zero_distance: Union[float, Distance]) -> Angular:
        """Sets shot.weapon.zero_elevation so that it hits a target at zero_distance.
        :param shot: Shot instance from which we take a zero
        :param zero_distance: Look-distance to "zero," which is point we want to hit.
        """
        shot.weapon.zero_elevation = self.barrel_elevation_for_target(shot, zero_distance)
        return shot.weapon.zero_elevation

    def fire(self, shot: Shot, trajectory_range: Union[float, Distance],
             trajectory_step: Union[float, Distance] = 0,
             extra_data: bool = False,
             time_step: float = 0.0) -> HitResult:
        """Calculates trajectory
        :param shot: shot parameters (initial position and barrel angle)
        :param trajectory_range: Downrange distance at which to stop computing trajectory
        :param trajectory_step: step between trajectory points to record
        :param extra_data: True => store TrajectoryData for every calculation step;
            False => store TrajectoryData only for each trajectory_step
        :param time_step: (seconds) If > 0 then record TrajectoryData at least this frequently
        """
        step: Distance
        trajectory_range = PreferredUnits.distance(trajectory_range)
        if not trajectory_step:
            trajectory_step = trajectory_range.raw_value / 10.0
            step = Distance.Inch(trajectory_step)
        else:
            step = PreferredUnits.distance(trajectory_step)
        data = self._calc.trajectory(shot, trajectory_range, step, extra_data, time_step)
        return HitResult(shot, data, extra_data)


__all__ = ('EulerCalculator',)



#!/usr/bin/env python
"""setup.py script for CyEulerBallistic library"""
from setuptools import setup, Extension
try:
    from Cython.Build import cythonize
except ImportError:
    import sys
    # use this command to skip building wheel and compiling modules on unsupported platforms
    # pip install --no-build-isolation --no-binary :all: py-ballisticcalc.exts
    setup()
    sys.exit(0)  # Stop installation


compiler_directives = {
    "language_level": 3,
    "embedsignature": True,
    "cdivision": True,
    "binding": False,
    "c_api_binop_methods": True,
    "warn.undeclared": True,
    "warn.unreachable": True,
    "warn.maybe_uninitialized": True,
    "warn.unused": True,
    "warn.unused_arg": True,
    "warn.multiple_declarators": True,
    "show_performance_hints": True,
}

extension_names = [
    "vector",
    "trajectory_calc",
    "trajectory_data",
    "cy_euler",
]

extensions = [
    Extension('CyEulerBallistic.' + name,
              ['CyEulerBallistic/' + name + '.pyx'])
    for name in extension_names
]

extensions = cythonize(extensions,
                       compiler_directives=compiler_directives,
                       annotate=True)

setup(ext_modules=extensions)



# type: ignore
from py_ballisticcalc.conditions import Shot
from py_ballisticcalc.drag_model import DragDataPoint
from py_ballisticcalc.unit import Angular, Distance
from py_ballisticcalc.trajectory_calc import Config
from py_ballisticcalc.trajectory_data import TrajectoryData

__all__ = ['EulerTrajectoryCalc']

class EulerTrajectoryCalc:
    def __init__(self, _config: Config) -> None: ...
    @property
    def table_data(self) -> list[DragDataPoint]: ...
    def get_calc_step(self, step: float = 0) -> float: ...
    def trajectory(self, shot_info: Shot, max_range: Distance, dist_step: Distance, extra_data: bool = False, time_step: float = 0.0) -> list[TrajectoryData]: ...
    def zero_angle(self, shot_info: Shot, distance: Distance, time_step: float = 0.0) -> Angular: ...
